package org.xbib.event.log;

import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.List;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.zip.GZIPOutputStream;

public class RotatingFileWriter extends Writer implements Rotatable {

    private static final Logger logger = Logger.getLogger(RotatingFileWriter.class.getName());

    private final RotationConfig config;

    private final List<RotationPolicy> writeSensitivePolicies;

    private volatile CharCountingWriter charCountingWriter;

    public RotatingFileWriter(RotationConfig config) throws IOException {
        this.config = config;
        Objects.requireNonNull(config.getScheduler());
        this.writeSensitivePolicies = config.getPolicies().stream().filter(RotationPolicy::isWriteSensitive).collect(Collectors.toList());
        this.charCountingWriter = open(null, config.getClock().now());
        startPolicies();
    }

    @Override
    public void rotate(RotationPolicy policy, Instant instant) throws IOException {
        Objects.requireNonNull(instant, "instant");
        RotationListener listener = config.getListener();
        listener.onTrigger(policy, instant);
        charCountingWriter.flush();
        if (Files.size(config.getPath()) == 0) {
            logger.log(Level.FINE, MessageFormat.format("empty file, skipping rotation {path={}}", config.getPath()));
            return;
        }
        listener.onClose(policy, instant);
        charCountingWriter.close();
        Path rotated = config.getFilePattern().create(instant);
        logger.log(Level.FINE, MessageFormat.format("renaming {file={0}, rotated={1}}", config.getPath(), rotated));
        Files.move(config.getPath(), rotated);
        logger.log(Level.FINE, MessageFormat.format("re-opening {path={0}}", config.getPath()));
        charCountingWriter = open(policy, instant);
        if (config.isCompress()) {
            compress(policy, instant, rotated, listener);
            return;
        }
        listener.onSuccess(policy, instant, rotated);
    }

    @Override
    public void onException(RotationPolicy policy, Instant instant, Path path, Throwable throwable) {
        config.getListener().onFailure(policy, instant, path, throwable);
    }

    private void compress(RotationPolicy policy, Instant instant, Path rotated, RotationListener listener) {
        config.getScheduler().execute(() -> {
            Path compressed = Paths.get(rotated + ".gz");
            logger.log(Level.FINE, MessageFormat.format("compressing {rotated={}, compressed={}}", rotated, compressed));
            try (InputStream inputStream = Files.newInputStream(rotated);
                 GZIPOutputStream gzipOutputStream = new GZIPOutputStream(Files.newOutputStream(compressed))) {
                inputStream.transferTo(gzipOutputStream);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                listener.onFailure(policy, instant, compressed, e);
                return;
            }
            logger.log(Level.FINE, MessageFormat.format("deleting old file {rotated={}}", rotated));
            try {
                Files.delete(rotated);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                listener.onFailure(policy, instant, compressed, e);
                return;
            }
            listener.onSuccess(policy, instant, compressed);
        });
    }

    @Override
    public RotationConfig getConfig() {
        return config;
    }

    @Override
    public synchronized void write(int b) throws IOException {
        long byteCount = charCountingWriter.size() + 1;
        notifyWriteSensitivePolicies(byteCount);
        charCountingWriter.write(b);
    }

    @Override
    public void write(char[] cbuf, int off, int len) throws IOException {
        long byteCount = charCountingWriter.size() + len;
        notifyWriteSensitivePolicies(byteCount);
        charCountingWriter.write(cbuf, off, len);
    }

    @Override
    public synchronized void flush() throws IOException {
        charCountingWriter.flush();
    }

    @Override
    public synchronized void close() throws IOException {
        config.getListener().onClose(null, config.getClock().now());
        if (charCountingWriter != null) {
            charCountingWriter.close();
            charCountingWriter = null;
        }
    }

    private void startPolicies() {
        for (RotationPolicy policy : config.getPolicies()) {
            policy.start(this);
        }
    }

    private CharCountingWriter open(RotationPolicy policy, Instant instant) throws IOException {
        Writer writer = config.isAppend() && Files.exists(config.getPath())?
                Files.newBufferedWriter(config.getPath(), StandardOpenOption.APPEND) :
                Files.newBufferedWriter(config.getPath());
        config.getListener().onOpen(policy, instant);
        long size = config.isAppend() ? Files.size(config.getPath()) : 0;
        return new CharCountingWriter(writer, size);
    }

    private void notifyWriteSensitivePolicies(long byteCount) {
        for (RotationPolicy writeSensitivePolicy : writeSensitivePolicies) {
            writeSensitivePolicy.acceptWrite(byteCount);
        }
    }

    @Override
    public String toString() {
        return "RotatingFileOutputStream " + config.getPath();
    }

}
