package org.xbib.event.path;

import org.xbib.event.Event;
import org.xbib.event.common.AbstractEventManagerService;
import org.xbib.event.common.EventManager;
import org.xbib.event.common.EventManagerService;
import org.xbib.settings.Settings;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;

public class FileFollowEventManagerService extends AbstractEventManagerService implements EventManagerService {

    private static final Logger logger = Logger.getLogger(FileFollowEventManagerService.class.getName());

    private EventManager eventManager;

    private final Map<Future<?>, FileFollowEventService> services;

    public FileFollowEventManagerService() {
        super();
        this.services = new LinkedHashMap<>();
    }

    @Override
    public FileFollowEventManagerService init(EventManager eventManager) {
        this.eventManager = eventManager;
        ExecutorService executorService = eventManager.getExecutorService();
        for (Map.Entry<String, Settings> entry : eventManager.getSettings().getGroups("event.filefollow").entrySet()) {
            String name = entry.getKey();
            Settings definition = entry.getValue();
            if (definition.getAsBoolean("enabled", true)) {
                String type = definition.get("type", "filefollow");
                String baseStr = definition.get("base");
                Objects.requireNonNull(baseStr);
                String patternStr = definition.get("pattern");
                Objects.requireNonNull(patternStr);
                try {
                    Path base = Paths.get(baseStr);
                    Pattern pattern = Pattern.compile(patternStr);
                    FileFollowEventService fileFollowEventService = new FileFollowEventService(this,
                            type, name, base, pattern);
                    Future<?> future = executorService.submit(fileFollowEventService);
                    services.put(future, fileFollowEventService);
                    logger.log(Level.INFO, "file follow service " + name +
                            " with type " + type +
                            " with base " + base + " and pattern " + pattern + " added");
                } catch (Exception e) {
                    logger.log(Level.SEVERE, "unable to create file follow service " +name + ", reason " + e.getMessage(), e);
                }
            }
        }
        logger.log(Level.INFO, "after init: event services = " + services);
        return this;
    }

    @Override
    public void shutdown() throws IOException {
        for (Map.Entry<Future<?>, FileFollowEventService> entry : services.entrySet()) {
            try {
                entry.getValue().setKeepWatching(false);
                entry.getKey().cancel(true);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    public void publish(Event event) {
        eventManager.publish(event);
    }
}
