package org.xbib.event.path;

import org.xbib.event.Event;
import org.xbib.event.common.EventManager;

import java.io.Closeable;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.SeekableByteChannel;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Stream;

public class FileFollowEventService implements Callable<Integer>, Closeable {

    private static final Logger logger = Logger.getLogger(FileFollowEventService.class.getName());

    private final FileFollowEventManagerService fileFollowEventManagerService;

    private final String eventType;

    private final String name;

    private final Path base;

    private final Pattern pattern;

    private final WatchService watchService;

    private final Map<Path, Long> fileSizes;

    private volatile boolean keepWatching;

    public FileFollowEventService(FileFollowEventManagerService fileFollowEventManagerService,
                                  String eventType,
                                  String name,
                                  Path base,
                                  Pattern pattern) throws IOException {
        this.fileFollowEventManagerService = fileFollowEventManagerService;
        this.eventType = eventType;
        this.name = name;
        this.base = base;
        this.pattern = pattern;
        FileSystem fileSystem = base.getFileSystem();
        this.watchService = fileSystem.newWatchService();
        WatchEvent.Kind<?>[] kinds = new WatchEvent.Kind<?>[1];
        kinds[0] = StandardWatchEventKinds.ENTRY_MODIFY;
        base.register(watchService, kinds);
        this.fileSizes = new LinkedHashMap<>();
        // initialize the sizes of existing files, so we can safely create first event
        setFileSizes(base, pattern);
        this.keepWatching = true;
    }

    public void setKeepWatching(boolean keepWatching) {
        this.keepWatching = keepWatching;
    }

    @SuppressWarnings("unchecked")
    @Override
    public Integer call() {
        while (keepWatching) {
            WatchKey key = null;
            try {
                key = watchService.take();
                for (WatchEvent<?> watchEvent : key.pollEvents()) {
                    WatchEvent<Path> pathWatchEvent = (WatchEvent<Path>) watchEvent;
                    Path path = pathWatchEvent.context();
                    Matcher matcher = pattern.matcher(path.toString());
                    if (!matcher.matches()) {
                        logger.log(Level.FINE, "no match of " + path + " to pattern " + pattern);
                        continue;
                    }
                    Path p = base.resolve(path);
                    try (SeekableByteChannel channel = Files.newByteChannel(p, StandardOpenOption.READ)) {
                        long lastSize = fileSizes.getOrDefault(p, 0L);
                        long currentSize = p.toFile().length();
                        fileSizes.put(p, currentSize);
                        String content = readRange(channel, lastSize, currentSize);
                        // split file content by line
                        // this prevents swalloed events and allows pattern matching without preprocessing in worker
                        for (String line : content.split("\n")) {
                            if (fileFollowEventManagerService.getSuspended().contains(name)) {
                                logger.log(Level.WARNING, name + " is suspended");
                            } else {
                                Event event = EventManager.eventBuilder()
                                        .setType(eventType)
                                        .setCode(base.toString())
                                        .setPath(path)
                                        .setMessage(line)
                                        .build();
                                fileFollowEventManagerService.publish(event);
                            }
                        }
                    }
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            } finally {
                if (key != null) {
                    key.reset();
                }
            }
        }
        return 0;
    }

    @Override
    public void close() throws IOException {
        watchService.close();
    }

    private static String readRange(SeekableByteChannel fileChannel, long from, long to) throws IOException {
        int delta = (int) (to - from);
        if (delta <= 0) {
            return "";
        }
        ByteBuffer byteBuffer = ByteBuffer.allocate(delta);
        fileChannel.position(from);
        int numRead = fileChannel.read(byteBuffer);
        byteBuffer.flip();
        CharBuffer charBuffer = StandardCharsets.UTF_8.decode(byteBuffer);
        byteBuffer.clear();
        if (numRead <= 0) {
            throw new IOException("numRead less or equal to 0");
        }
        return charBuffer.toString();
    }

    private void setFileSizes(Path base, Pattern pattern) throws IOException {
        if (!Files.exists(base)) {
            return;
        }
        try (Stream<Path> path = Files.walk(base)) {
            path.forEach(p -> {
                Matcher matcher = pattern.matcher(p.toString());
                if (matcher.matches()) {
                    try {
                        fileSizes.put(p, Files.size(p));
                    } catch (IOException e) {
                        throw new UncheckedIOException(e);
                    }
                }
            });
        }
    }
}
