package org.xbib.event.timer;

import org.xbib.event.Event;
import org.xbib.event.common.PayloadImpl;
import org.xbib.event.common.EventManager;
import org.xbib.event.persistence.PersistenceStore;

import java.io.Closeable;
import java.io.IOException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

class TimerEventService implements Closeable {

    private static final Logger logger = Logger.getLogger(TimerEventService.class.getName());

    private final TimerEventManagerService timerEventManagerService;

    private final String name;

    private final ZoneId zoneId;

    private final PersistenceStore<String, Object> persistenceStore;

    private final Timer timer;

    public TimerEventService(TimerEventManagerService timerEventManagerService,
                             String name,
                             ZoneId zoneId,
                             PersistenceStore<String, Object> persistenceStore) throws IOException {
        this.timerEventManagerService = timerEventManagerService;
        this.name = name;
        this.zoneId = zoneId;
        this.persistenceStore = persistenceStore;
        this.timer = new Timer();
        int tasknum = reschedule();
        logger.log(Level.INFO, "timer event service " + name + " loaded and rescheduled, " +  tasknum + " timer tasks");
    }

    void schedule(Instant instant, PayloadImpl payload) throws IOException {
        String scheduled = instant.atZone(zoneId).format(DateTimeFormatter.ISO_DATE_TIME);
        payload.put("scheduled", scheduled);
        TimerEventTask timerEventTask = new TimerEventTask(payload);
        timer.schedule(timerEventTask, Date.from(instant));
        logger.log(Level.INFO, "new task " + payload + " added, scheduled at " + scheduled);
    }

    @SuppressWarnings("unchecked")
    int reschedule() throws IOException {
        logger.log(Level.INFO, "rescheduling timer events");
        persistenceStore.load();
        List<Map<String, Object>> tasks = (List<Map<String, Object>>) persistenceStore.getOrDefault("tasks", new ArrayList<>());
        persistenceStore.clear();
        persistenceStore.commit();
        for (Map<String, Object> task : tasks) {
            PayloadImpl payload = new PayloadImpl(task);
            ZonedDateTime scheduledDate = ZonedDateTime.parse((String) task.get("scheduled"), DateTimeFormatter.ISO_DATE_TIME);
            if (scheduledDate.isBefore(ZonedDateTime.now())) {
                logger.log(Level.WARNING, "scheduled timer task " + task + " date already passed");
            } else {
                schedule(scheduledDate.toInstant(), payload);
            }
        }
        tasks = (List<Map<String, Object>>) persistenceStore.getOrDefault("tasks", new ArrayList<>());
        logger.log(Level.INFO, "rescheduling timer events complete: " + tasks.size() + " events");
        return tasks.size();
    }

    @Override
    public void close() throws IOException {
        timer.cancel();
        persistenceStore.commit();
    }

    void purge() {
        timer.purge();
    }

    public class TimerEventTask extends TimerTask {

        private final PayloadImpl payload;

        public TimerEventTask(PayloadImpl payload) throws IOException {
            this.payload = payload;
            persistenceStore.insert("tasks", this.payload);
        }

        @Override
        public void run() {
            try {
                if (timerEventManagerService.getSuspended().contains(name)) {
                    logger.log(Level.FINE, "timer event " + name + " suspended");
                    return;
                }
                Event timerEvent = EventManager.eventBuilder()
                        .setType("timer")
                        .setCode(name)
                        .setPayload(payload)
                        .build();
                timerEventManagerService.publish(timerEvent);
                logger.log(Level.FINE, "persistence before remove: " + persistenceStore.get("tasks"));
                if (persistenceStore.remove("tasks", this.payload)) {
                    logger.log(Level.FINE, "removal done");
                }
                logger.log(Level.FINE, "persistence after remove: " + persistenceStore.get("tasks"));
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        @Override
        public String toString() {
            return payload.toString();
        }

        @Override
        public int hashCode() {
            return payload.hashCode();
        }

        @Override
        public boolean equals(Object object) {
            return object instanceof TimerEventTask && Objects.equals(payload, ((TimerEventTask) object).payload);
        }
    }
}
