package org.xbib.event.common;

import org.junit.jupiter.api.Test;
import org.xbib.event.Event;
import org.xbib.event.EventConsumer;
import org.xbib.event.bus.Subscribe;
import org.xbib.settings.Settings;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

public class GenericEventManagerTest {

    private static final Logger logger = Logger.getLogger(GenericEventManagerTest.class.getName());

    @Test
    void testSimpleGenericEvent() throws InterruptedException {
        TestEventConsumer consumer = new TestEventConsumer();
        EventManager eventManager = EventManager.builder()
                .register(consumer)
                .build();
        Event event = EventManager.eventBuilder()
                .setType("generic")
                .setListener(e -> logger.log(Level.INFO, "received event " + e))
                .build();
        eventManager.publish(event);
        // we must wait for a certain time because we do not use a future
        Thread.sleep(500L);
        assertEquals(1, consumer.getCount());
    }

    @Test
    void testGenericEventWithWaitForSubscriber() throws InterruptedException, ExecutionException, TimeoutException {
        TestEventConsumer consumer = new TestEventConsumer();
        EventManager eventManager = EventManager.builder()
                .register(consumer)
                .build();
        CompletableFuture<Event> future = new CompletableFuture<>();
        Event event = EventManager.eventBuilder()
                .setType("generic")
                .setListener(e -> {
                    logger.log(Level.INFO, "received event " + e);
                    future.complete(e);
                })
                .build();
        eventManager.publish(event);
        Event e = future.get(1000L, TimeUnit.MILLISECONDS);
        assertNotNull(e);
    }

    @Test
    void testGenericEventWithWaitForAllConsumers() throws ExecutionException, InterruptedException, TimeoutException {
        Settings settings = Settings.settingsBuilder()
                .build();
        TestEventConsumer consumer1 = new TestEventConsumer();
        TestEventConsumer consumer2 = new TestEventConsumer();
        EventManager eventManager = EventManager.builder()
                .setSettings(settings)
                .register(consumer1)
                .register(consumer2)
                .build();
        CompletableFuture<Event> future = new CompletableFuture<>();
        Event event = EventManager.eventBuilder()
                .setType("generic")
                .setListener(e -> {
                    logger.log(Level.INFO, "received event " + e);
                    future.complete(e);
                })
                .build();
        eventManager.publish((GenericEventImpl) event, future);
        Event e = future.get(5L, TimeUnit.SECONDS);
        if (e != null) {
            logger.log(Level.INFO, "the event " + e + " was received by all consumers");
        }
        assertNotNull(e);
    }

    private static class TestEventConsumer implements EventConsumer {

        private int count;

        TestEventConsumer() {
        }

        @Subscribe
        public void onEvent(GenericEventImpl event) {
            event.received();
            count++;
        }

        public int getCount() {
            return count;
        }
    }
}
