package org.xbib.event.path;

import org.junit.jupiter.api.Test;
import org.xbib.datastructures.api.TimeValue;
import org.xbib.event.EventConsumer;
import org.xbib.event.PathEvent;
import org.xbib.event.bus.AllowConcurrentEvents;
import org.xbib.event.bus.Subscribe;
import org.xbib.event.common.EventManager;
import org.xbib.event.common.PathEventImpl;
import org.xbib.settings.Settings;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PathEventManagerTest {

    private static final Logger logger = Logger.getLogger(PathEventManagerTest.class.getName());

    @Test
    public void testPathEvents() throws IOException, InterruptedException {
        Path path = Files.createTempDirectory("testpath");
        TestPathEventConsumer consumer = new TestPathEventConsumer();
        Settings settings = Settings.settingsBuilder()
                .put("event.path.testpathevent.enabled", "true")
                .put("event.path.testpathevent.path", path.toString())
                .build();
        EventManager eventManager = EventManager.builder()
                .setSettings(settings)
                .register(consumer)
                .build();
        Path testTxt = path.resolve("incoming").resolve("test.txt");
        try (BufferedWriter bufferedWriter = Files.newBufferedWriter(testTxt)) {
            bufferedWriter.write("{\"Hello\":\"world\"}");
        }
        Thread.sleep(1000L);
        eventManager.shutdown();
        // extra destroy to clean up test
        eventManager.getPathEventManagerService().destroy();
    }

    @Test
    public void testExtendedPathEvent() throws Exception {
        Path path = Files.createTempDirectory("testpath");
        TestPathEventConsumer consumer = new TestPathEventConsumer();
        Settings settings = Settings.settingsBuilder()
                .put("event.path.testpathevent.enabled", "true")
                .put("event.path.testpathevent.type", "path-ext")
                .put("event.path.testpathevent.path", path.toString())
                .build();
        EventManager eventManager = EventManager.builder()
                .setSettings(settings)
                .register("path-ext", PathExtEvent.class)
                .register(consumer)
                .build();
        Path testTxt = path.resolve("incoming").resolve("test.txt");
        try (BufferedWriter bufferedWriter = Files.newBufferedWriter(testTxt)) {
            bufferedWriter.write("{\"Hello\":\"world\"}");
        }
        Thread.sleep(1000L);
        eventManager.shutdown();
        // extra destroy to clean up test
        eventManager.getPathEventManagerService().destroy();
    }

    @Test
    public void testPathEventByApi() throws Exception {
        TestPathExtEventConsumer consumer = new TestPathExtEventConsumer();
        EventManager eventManager = EventManager.builder()
                .register("path-ext", PathExtEvent.class)
                .register(consumer)
                .build();
        // create by API
        Path path = Files.createTempDirectory("testpath");
        eventManager.getPathEventManagerService().add("test", path, "path-ext", TimeValue.timeValueHours(72));
        // publish
        PathEventService pathEventService = eventManager.getPathEventManagerService().getPathEventService("test");
        eventManager.getPathEventManagerService()
                .publishJson(pathEventService, "key", Map.of("hello", "world"));
        // everything done
        Thread.sleep(1000L);
        eventManager.shutdown();
        // extra destroy to clean up test
        eventManager.getPathEventManagerService().destroy();
    }

    public static class PathExtEvent extends PathEventImpl implements PathEvent {

        public PathExtEvent(EventManager.EventBuilder builder) {
            super(builder);
            logger.log(Level.INFO, "I'm the path ext event");
        }
    }

    public static class TestPathEventConsumer implements EventConsumer {

        @Subscribe
        @AllowConcurrentEvents
        void onEvent(PathEvent event) {
            logger.log(Level.INFO, "received path event, path = " + event.getPath() +
                    " payload = " + event.getPayload());
        }
    }


    public static class TestPathExtEventConsumer implements EventConsumer {

        @Subscribe
        @AllowConcurrentEvents
        void onEvent(PathExtEvent event) {
            logger.log(Level.INFO, "received path ext event, path = " + event.getPath() +
                    " payload = " + event.getPayload());
        }
    }
}
