package org.xbib.event.pubsub;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.SubscriberExceptionContext;
import org.xbib.event.bus.SubscriberExceptionHandler;

public class MessageBroker<E> {

    private static final Logger logger = Logger.getLogger(MessageBroker.class.getName());

    private final Builder builder;

    private final Collection<Topic<E>> topics;

    private MessageBroker(Builder builder) {
        this.builder = builder;
        this.topics = new ArrayList<>();
    }

    public void subscribe(Subscriber<E> subscriber, Topic<E> topic) {
        topic.add(subscriber);
        if (!topics.contains(topic)) {
            topics.add(topic);
        }
        builder.eventBus.register(subscriber);
    }

    public void unsubscribe(Subscriber<E> subscriber, Topic<E> topic) {
        topic.remove(subscriber);
        builder.eventBus.unregister(subscriber);
    }

    public void publish(E event) {
        builder.eventBus.post(event);
    }

    public static class Builder {

        private int threadCount;

        private ExecutorService executorService;

        private ClassLoader classLoader;

        private SubscriberExceptionHandler subscriberExceptionHandler;

        private AsyncEventBus eventBus;

        private Builder() {
            this.subscriberExceptionHandler = new EventManagerExceptionHandler();
        }

        public Builder setThreadCount(int threadCount) {
            this.threadCount = threadCount;
            return this;
        }

        public Builder setExecutorService(ExecutorService executorService) {
            this.executorService = executorService;
            return this;
        }

        public Builder setClassLoader(ClassLoader classLoader) {
            this.classLoader = classLoader;
            return this;
        }

        public Builder setEventBus(AsyncEventBus eventBus) {
            this.eventBus = eventBus;
            return this;
        }

        public Builder setSubscriberExceptionHandler(SubscriberExceptionHandler subscriberExceptionHandler) {
            this.subscriberExceptionHandler = subscriberExceptionHandler;
            return this;
        }

        public MessageBroker build() {
            if (executorService == null) {
                executorService = threadCount > 0 ?
                        Executors.newFixedThreadPool(threadCount) :
                        Executors.newCachedThreadPool();
            }
            if (classLoader == null) {
                classLoader = MessageBroker.class.getClassLoader();
            }
            if (eventBus == null) {
                eventBus = new AsyncEventBus(executorService, subscriberExceptionHandler);
            }
            return new MessageBroker(this);
        }
    }

    private static class EventManagerExceptionHandler implements SubscriberExceptionHandler {
        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            logger.log(Level.SEVERE, exception.getMessage(), exception);
        }
    }
}
